import struct
import logging

import numpy as np

from vaex.dataset_mmap import DatasetMemoryMapped


logger = logging.getLogger("vaex.astro.gadget")


def getinfo(filename, seek=None):
	"""Read header data from Gadget data file 'filename' with Gadget file
	type 'gtype'. Returns offsets of positions and velocities."""
	DESC = '=I4sII'                                # struct formatting string
	HEAD = '=I6I6dddii6iiiddddii6ii60xI'        # struct formatting string
	keys = ('Npart', 'Massarr', 'Time', 'Redshift', 'FlagSfr', 'FlagFeedback', 'Nall', 'FlagCooling', 'NumFiles', 'BoxSize', 'Omega0', 'OmegaLambda', 'HubbleParam', 'FlagAge', 'FlagMetals', 'NallHW', 'flag_entr_ics', 'filename')
	f = open(filename, 'rb')
	
	"""Detects Gadget file type (type 1 or 2; resp. without or with the 16
	byte block headers)."""
	firstbytes = struct.unpack('I',f.read(4))
	if firstbytes[0] == 8:
		gtype = 2
	else:
		gtype = 1
	if gtype == 2:
		f.seek(16)
	else:
	 	f.seek(0)
	if seek is not None:
		f.seek(seek)
	raw = struct.unpack(HEAD,f.read(264))[1:-1]
	values = (raw[:6], raw[6:12]) + raw[12:16] + (raw[16:22],) + raw[22:30] + (raw[30:36], raw[36], filename)
	header = dict(list(zip(keys, values)))
	f.close()
	
	
	if gtype == 2:
		posoffset = (2*16 + (8 + 256))
	else:
		posoffset = (8 + 256)
	
	Npart = sum(header['Npart'])
	if gtype == 2:
		veloffset = 3*16 + (8 + 256) + (8 + 3*4*Npart)
	else:
		veloffset= (8 + 256) + (8 + 3*4*Npart)
	return Npart, posoffset+4, veloffset+4, header


class MemoryMappedGadget(DatasetMemoryMapped):
	def __init__(self, filename, fs_options={}):
		super(MemoryMappedGadget, self).__init__(filename)
		#h5file = h5py.File(self.filename)
		length, posoffset, veloffset, header = getinfo(filename)
		self._add("x", posoffset, length, dtype=np.float32, stride=3)
		self._add("y", posoffset+4, length, dtype=np.float32, stride=3)
		self._add("z", posoffset+8, length, dtype=np.float32, stride=3)

		self._add("vx", veloffset, length, dtype=np.float32, stride=3)
		self._add("vy", veloffset+4, length, dtype=np.float32, stride=3)
		self._add("vz", veloffset+8, length, dtype=np.float32, stride=3)
		self._freeze()

	def _add(self, name, offset, length, dtype, stride):
		ar = self._map_array(offset, length, dtype)
		ar = ar[::stride]
		self.add_column(name, ar)


	@classmethod
	def can_open(cls, path, *args, **kwargs):
		with open(path, 'rb') as f:
			try:
				first_words = struct.unpack('4I',f.read(4*4))
			except struct.error:
				return False
			if first_words[0] == 8 and first_words[2] == 8 and first_words[3] == 256:
				logger.debug('gadget file SnapFormat=2 detected')
				return True
			elif first_words[0] == 256:
				f.seek(256+4)
				if struct.unpack('I',f.read(4))[0] == 256:
					logger.debug('gadget file SnapFormat=1 detected')
					return True
		return False





import numpy as np


from vaex.dataset import DatasetArrays


class DatasetTap(DatasetArrays):
    class TapColumn(object):
        def __init__(self, tap_dataset, column_name, column_type, ucd):
            self.tap_dataset = tap_dataset
            self.column_name = column_name
            self.column_type = column_type
            self.ucd = ucd
            self.alpha_min = 0
            length = len(tap_dataset)
            steps = length/1e6 # try to do it in chunks
            self.alpha_step = 360/steps
            self.alpha_max = self.alpha_min + self.alpha_step
            logger.debug("stepping in alpha %f" % self.alpha_step)
            self.data = []
            self.offset = 0
            self.shape = (length,)
            self.dtype = DatasetTap.type_map[self.column_type]().dtype
            self.left_over_chunk = None
            self.rows_left = length
            import tempfile
            self.download_file = tempfile.mktemp(".vot")

        def __getitem__(self, slice):
            start, stop, step = slice.start, slice.stop, slice.step
            required_length = stop - start
            assert start >= self.offset
            chunk_data = self.left_over_chunk
            enough = False if chunk_data is None else len(chunk_data) >= required_length
            if chunk_data is not None:
                logger.debug("start %s offset %s chunk length %s", start, self.offset, len(chunk_data))
                #assert len(chunk_data) == start - self.offset
            if enough:
                logger.debug("we can skip the query, already have results from previous query")
            while not enough:
                adql_query = "SELECT {column_name} FROM {table_name} WHERE alpha >= {alpha_min} AND alpha < {alpha_max} ORDER BY alpha ASC"\
                    .format(column_name=self.column_name, table_name=self.tap_dataset.table_name, alpha_min=self.alpha_min, alpha_max=self.alpha_max)
                logger.debug("executing: %s" % adql_query)
                logger.debug("executing: %s" % adql_query.replace(" ", "+"))


                url = self.tap_dataset.tap_url + "/sync?REQUEST=doQuery&LANG=ADQL&MAXREC=10000000&FORMAT=votable&QUERY=" +adql_query.replace(" ", "+")
                import urllib2
                response = urllib2.urlopen(url)
                with open(self.download_file, "w") as f:
                    f.write(response.read())
                votable = astropy.io.votable.parse(self.download_file)
                data = votable.get_first_table().array[self.column_name].data
                # TODO: respect masked array
                #table = astropy.table.Table.read(url, format="votable") #, show_progress=False)
                #data = table[self.column_name].data.data.data
                logger.debug("new chunk is of lenght %d", len(data))
                self.rows_left -= len(data)
                logger.debug("rows left %d", self.rows_left)
                if chunk_data is None:
                    chunk_data = data
                else:
                    chunk_data = np.concatenate([chunk_data, data])
                if len(chunk_data) >= required_length:
                    enough = True
                logger.debug("total chunk is of lenght %d, enough: %s", len(chunk_data), enough)
                self.alpha_min += self.alpha_step
                self.alpha_max += self.alpha_step


            result, self.left_over_chunk = chunk_data[:required_length], chunk_data[required_length:]
            #print(result)
            logger.debug("left over is of length %d", len(self.left_over_chunk))
            return result #np.zeros(N, dtype=self.dtype)



    type_map = {
        'REAL':np.float32,
        'SMALLINT':np.int32,
        'DOUBLE':np.float64,
        'BIGINT':np.int64,
        'INTEGER':np.int32,
        'BOOLEAN':np.bool8
    }
    #not supported types yet 'VARCHAR',', u'BOOLEAN', u'INTEGER', u'CHAR
    def __init__(self, tap_url="http://gaia.esac.esa.int/tap-server/tap/g10_smc", table_name=None):
        logger.debug("tap url: %r", tap_url)
        self.tap_url = tap_url
        self.table_name = table_name
        if table_name is None: # let us try to infer the table name
            if tap_url.endswith("tap") or tap_url.endswith("tap/"):
                pass # this mean we really didn't provide one
            else:
                index = tap_url.rfind("tap/")
                if index != -1:
                    self.tap_url, self.table_name = tap_url[:index+4], self.tap_url[index+4:]
                    logger.debug("inferred url is %s, and table name is %s", self.tap_url, self.table_name)

        if self.tap_url.startswith("tap+"): # remove tap+ part from tap+http(s), only keep http(s) part
            self.tap_url = self.tap_url[len("tap+"):]
        import requests
        super(DatasetTap, self).__init__(self.table_name)
        self.req = requests.request("get", self.tap_url+"/tables/")
        self.path = "tap+" +self.tap_url + "/" + table_name

        #print dir(self.req)
        from bs4 import BeautifulSoup
            #self.soup = BeautifulSoup(req.response)
        tables = BeautifulSoup(self.req.content, 'xml')
        self.tap_tables = collections.OrderedDict()
        for table in tables.find_all("table"):
            #print table.find("name").string, table.description.string, table["gaiatap:size"]
            table_name = unicode(table.find("name").string)
            table_size = int(table["esatapplus:size"])
            #print table_name, table_size
            logger.debug("tap table %r ", table_name)
            columns = []
            for column in table.find_all("column"):
                column_name = unicode(column.find("name").string)
                column_type = unicode(column.dataType.string)
                ucd = column.ucd.string if column.ucd else None
                unit = column.unit.string if column.unit else None
                description = column.description.string if column.description else None
                #print "\t", column_name, column_type, ucd
                #types.add()
                columns.append((column_name, column_type, ucd, unit, description))
            self.tap_tables[table_name] = (table_size, columns)
        if not self.tap_tables:
            raise ValueError("no tables or wrong url")
        for name, (table_size, columns) in self.tap_tables.items():
            logger.debug("table %s has length %d", name, table_size)
        self._full_length, self._tap_columns = self.tap_tables[self.table_name]
        self._length = self._full_length
        logger.debug("selected table table %s has length %d", self.table_name, self._full_length)
        #self.column_names = []
        #self.columns = collections.OrderedDict()
        for column_name, column_type, ucd, unit, description in self._tap_columns:
            logger.debug("  column %s has type %s and ucd %s, unit %s and description %s", column_name, column_type, ucd, unit, description)
            if column_type in self.type_map.keys():
                self.column_names.append(column_name)
                if ucd:
                    self.ucds[column_name] = ucd
                if unit:
                    self.units[column_name] = unit
                if description:
                    self.descriptions[column_name] = description
                self.columns[column_name] = self.TapColumn(self, column_name, column_type, ucd)
            else:
                logger.warning("  type of column %s is not supported, it will be skipped", column_name)


    @classmethod
    def open(cls, path, *args, **kwargs):
        return cls(path, *args, **kwargs)

    @classmethod
    def quick_test(cls, path, *args, **kwargs):
        return False

    @classmethod
    def can_open(cls, path, *args, **kwargs):
        can_open = False
        url = None
        try:
            url = urlparse(path)
        except:
            return False
        if url.scheme:
            if url.scheme.startswith("tap+http"): # will also catch https
                can_open = True
        logger.debug("%r can open: %r"  %(cls.__name__, can_open))
        return can_open



import astropy.io.votable

import vaex
from vaex.dataset import DatasetFile
from vaex.dataset_misc import _try_unit

class VOTable(DatasetFile):
	def __init__(self, filename, fs_options):
		super().__init__(filename)
		self.ucds = {}
		self.units = {}
		self.filename = filename
		self.path = filename
		votable = astropy.io.votable.parse(self.filename)

		self.first_table = votable.get_first_table()
		self.description = self.first_table.description

		for field in self.first_table.fields:
			name = field.name
			data = self.first_table.array[name]
			type = self.first_table.array[name].dtype
			clean_name = name
			if field.ucd:
				self.ucds[clean_name] = field.ucd
			if field.unit:
				unit = _try_unit(field.unit)
				if unit:
					self.units[clean_name] = unit
			if field.description:
				self.descriptions[clean_name] = field.description
			if type.kind in "fiubSU": # only store float and int and boolean
				self.add_column(clean_name, data) #self.first_table.array[name].data)
			if type.kind == "O":
				print("column %r is of unsupported object type , will try to convert it to string" % (name,))
				try:
					data = data.astype("S")
					self.add_column(name, data)
				except Exception as e:
					print("Giving up column %s, error: %r" (name, e))
			#if type.kind in ["S"]:
			#	self.add_column(clean_name, self.first_table.array[name].data)
		self._freeze()

	@classmethod
	def can_open(cls, path, *args, **kwargs):
		can_open = vaex.file.stringyfy(path).endswith(".vot")
		return can_open

	def close(self):
		pass



import numpy as np

from vaex.dataset import DatasetArrays
from vaex.dataset_misc import _try_unit

class DatasetAstropyTable(DatasetArrays):
    def __init__(self, filename=None, format=None, table=None, **kwargs):
        self.ucds = {}
        self.units = {}
        columns = {}
        if table is None:
            self.filename = filename
            self.format = format
            self.read_table()
        else:
            self.description = table.meta.get("description")
            self.table = table

        for i in range(len(self.table.dtype)):
            name = self.table.dtype.names[i]
            column = self.table[name]
            type = self.table.dtype[i]
            if type.kind in "fiuSU": # only store float and int
                masked_array = self.table[name].data
                if "ucd" in column._meta:
                    self.ucds[name] = column._meta["ucd"]
                if column.unit:
                    unit = _try_unit(column.unit)
                    if unit:
                        self.units[name] = unit
                if column.description:
                    self.descriptions[name] = column.description
                if hasattr(masked_array, "mask"):
                    if type.kind in ["f"]:
                        masked_array.data[masked_array.mask] = np.nan
                    if type.kind in ["i"]:
                        masked_array.data[masked_array.mask] = 0
                columns[name] = self.table[name].data
            if type.kind in ["SU"]:
                columns[name] = self.table[name].data

        super().__init__(columns)

    def read_table(self):
        self.table = astropy.table.Table.read(self.filename, format=self.format, **kwargs)



import logging
import math
import os
import sys

import astropy.io.fits
import astropy.io.fits as fits
import numpy as np

import vaex.dataset
from vaex.dataset_mmap import DatasetMemoryMapped
import vaex.export


logger = logging.getLogger("vaex.astro.fits")


class FitsBinTable(DatasetMemoryMapped):
    def __init__(self, filename, write=False, fs_options={}):
        super(FitsBinTable, self).__init__(filename, write=write)
        self.ucds = {}
        self.units = {}
        with fits.open(filename) as fitsfile:
            for table in fitsfile:
                if isinstance(table, fits.BinTableHDU):
                    table_offset = table._data_offset
                    #import pdb
                    #pdb.set_trace()
                    if table.columns[0].dim is not None: # for sure not a colfits
                        dim = eval(table.columns[0].dim) # TODO: can we not do an eval here? not so safe
                        if len(dim) == 2 and dim[0] <= dim[1]: # we have colfits format
                            logger.debug("colfits file!")
                            offset = table_offset
                            for i in range(len(table.columns)):
                                column = table.columns[i]
                                cannot_handle = False
                                column_name = column.name.strip()
                                self._get_column_meta_data(table, column_name, column, i)


                                # flatlength == length * arraylength
                                flatlength, fitstype = int(column.format[:-1]),column.format[-1]
                                arraylength, length = arrayshape = eval(column.dim)

                                # numpy dtype code, like f8, i4
                                dtypecode = astropy.io.fits.column.FITS2NUMPY[fitstype]


                                dtype = np.dtype((">" +dtypecode, arraylength))
                                if 0:
                                    if arraylength > 1:
                                        dtype = np.dtype((">" +dtypecode, arraylength))
                                    else:
                                        if dtypecode == "a": # I think numpy needs by default a length 1
                                            dtype = np.dtype(dtypecode + "1")
                                        else:
                                            dtype = np.dtype(">" +dtypecode)
                                    #	bytessize = 8

                                bytessize = dtype.itemsize
                                logger.debug("%r", (column.name, dtype, column.format, column.dim, length, bytessize, arraylength))
                                if (flatlength > 0): # and dtypecode != "a": # TODO: support strings
                                    if dtypecode == "a": # for ascii, we need to add the length again..
                                        dtypecode += str(arraylength)
                                    logger.debug("column type: %r", (column.name, offset, dtype, length, column.format, column.dim))
                                    if arraylength == 1 or dtypecode[0] == "a":
                                        ar = self._map_array(offset=offset, dtype=dtype, length=length)
                                        self.add_column(column_name, ar)
                                    else:
                                        for i in range(arraylength):
                                            name = column_name+"_" +str(i)
                                            self.addColumn(name, offset=offset+bytessize*i//arraylength, dtype=">" +dtypecode, length=length, stride=arraylength)
                                if flatlength > 0: # flatlength can be
                                    offset += bytessize * length
                                self._check_null(table, column_name, column, i)

                    else:
                        logger.debug("adding table: %r" % table)
                        for i, column in enumerate(table.columns):
                            array = column.array[:]
                            array = column.array[:] # 2nd time it will be a real np array
                            #import pdb
                            #pdb.set_trace()
                            if array.dtype.kind in "fiubSU":
                                column_name = _python_save_name(column.name, used=self.columns.keys())
                                self.addColumn(column_name, array=array)
                                self._get_column_meta_data(table, column_name, column, i)
                                self._check_null(table, column_name, column, i)
            self._try_votable(fitsfile[0])

        self._freeze()
        # self.update_meta()
        # self.update_virtual_meta()

    def _check_null(self, table, column_name, column, i):
        null_name = "TNULL%d" % (i+1)
        if null_name in table.header:
            mask_value = table.header[null_name]
            array = self._columns[column_name]
            mask = array == mask_value
            self._columns[column_name] = np.ma.masked_array(array, mask)

    def _try_votable(self, table):
        try:
            from io import BytesIO as StringIO
        except:
            from StringIO import StringIO
        if table.data is None:
            return
        vodata = table.data.tostring()
        if vodata.startswith(b"<?xml"):
            f = StringIO()
            f.write(vodata)
            votable = astropy.io.votable.parse(f)
            first_table = votable.get_first_table()
            used_names = []
            for field in first_table.fields:
                name = field.name.strip()
                clean_name = _python_save_name(name, used=used_names)
                used_names.append(name)
                if field.ucd:
                    self.ucds[clean_name] = field.ucd
                unit = _try_unit(field.unit)
                if unit:
                    self.units[clean_name] = unit
                if unit is None and field.unit:
                    print("unit error for: %r", field.unit)
                self.descriptions[clean_name] = field.description
            self.description = first_table.description

    def _get_column_meta_data(self, table, column_name, column, i):
        ucd_header_name = "TUCD%d" % (i+1)
        if ucd_header_name in table.header:
            self.ucds[column_name] = table.header[ucd_header_name]
        if column.unit:
            try:
                unit = _try_unit(column.unit)
                if unit:
                    self.units[column_name] = unit
            except:
                logger.exception("could not understand unit: %s" % column.unit)
        else: # we may want to try ourselves
            unit_header_name = "TUNIT%d" % (i+1)
            if unit_header_name in table.header:
                unit_str = table.header[unit_header_name]
                unit = _try_unit(unit_str)
                if unit:
                    self.unit[column_name] = unit
        #unit_header_name = "TUCD%d" % (i+1)
        #if ucd_header_name in table.header:
    @classmethod
    def can_open(cls, path, *args, **kwargs):
        return os.path.splitext(path)[1] == ".fits"

    @classmethod
    def get_options(cls, path):
        return [] # future: support multiple tables?

    @classmethod
    def option_to_args(cls, option):
        return []


def empty(filename, length, column_names, data_types, data_shapes, ucds, units, null_values={}):
	with open(filename, "wb") as f:
		logger.debug("preparing empty fits file: %s", filename)
		class Scope(object):
			pass

		def write(key, value, comment=""):
			first_part = "{key:8}= {value:20} / ".format(key=key, value=value)
			f.write(first_part.encode("ascii"))
			leftover = 80 - len(first_part)
			f.write(("{comment:"+str(leftover) +"}").format(comment=comment).encode("ascii"))
			logger.debug("at pos: %s", f.tell())
		def finish_header():
			f.write("{end:80}".format(end="END").encode("ascii"))
			offset = f.tell()
			bytes_over_padding = offset % 2880
			logger.debug(("bytes_over_padding: %s", bytes_over_padding))
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write((" "*padding).encode("ascii"))
		def finish_data():
			offset = f.tell()
			bytes_over_padding = offset % 2880
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write(("\0"*padding).encode("ascii"))

		byte_size = sum([length * type.itemsize for type in data_types])

		write("SIMPLE", "T", "file conforms to FITS standard")
		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 0, "number of array dimensions")
		finish_header()
		write("XTENSION", repr("BINTABLE"), "binary table extension")

		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 2, "number of array dimensions")
		write("NAXIS1", byte_size, "length of dim 1")
		write("NAXIS2", 1, "length of dim 2")
		write("PCOUNT", 0, "number of group parameters")
		write("GCOUNT", 1, "number of groups")

		write("TFIELDS", len(column_names), "number of columns")
		for i, (column_name, type, shape) in enumerate(zip(column_names, data_types, data_shapes)):
			i += 1 # 1 based index
			#column = dataset.columns[column_name]
			write("TTYPE%d" % i, repr(str(column_name)), "column name %i" % (i))
			numpy_type_name = type.descr[0][1][1:] # i4, f8 etc
			if numpy_type_name[0] == 'S':
				string_length = numpy_type_name[1:]
				fits_type = str(int(string_length)*length)+"A"
				logger.debug("type for %s: numpy=%r, fits=%r, string_length=%r length=%r", column_name, numpy_type_name, fits_type, string_length, length)
				# TODO: support rank1 arrays
				write("TFORM%d" % i , repr("{type}".format(type=fits_type)), "")
				write("TDIM%d" % i, repr("({string_length},{length})".format(string_length=string_length, length=length)), "")
			else:
				fits_type = astropy.io.fits.column.NUMPY2FITS[numpy_type_name]
				logger.debug("type for %s: numpy=%r, fits=%r", column_name, numpy_type_name, fits_type)
				# TODO: support rank1 arrays
				write("TFORM%d" % i , repr("{length}{type}".format(length=length, type=fits_type)), "")
				write("TDIM%d" % i, repr("(1,{length})".format(length=length)), "")
			ucd = ucds[i-1]
			if ucd:
				write("TUCD%d" % i, repr(str(ucd)))
			unit = units[i-1]
			if unit:
				write("TUNIT%d" % i, repr(str(unit)))
			if column_name in null_values:
				write("TNULL%d" % i, str(null_values[column_name]))

		finish_header()

		for i, (column_name, type, shape) in enumerate(zip(column_names, data_types, data_shapes)):
			byte_size = length * type.itemsize
			f.seek(f.tell() + byte_size)
		finish_data()

def write_colfits(dataset, path, selection=False):
	with open(path, "wb") as f:
		class Scope(object):
			pass

		vars = Scope()
		#vars.
		def write(key, value, comment=""):
			f.write("{key:8}= {value:20} / {comment:47}".format(key=key, value=value, comment=comment))
			print(("at pos", f.tell()))
		def finish_header():
			print((f.write("{end:80}".format(end="END"))))
			offset = f.tell()
			bytes_over_padding = offset % 2880
			print(("bytes_over_padding", bytes_over_padding))
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write(" "*padding)
		def finish_data():
			offset = f.tell()
			bytes_over_padding = offset % 2880
			if bytes_over_padding > 0:
				padding = 2880 - bytes_over_padding
				f.write("\0"*padding)


		write("SIMPLE", "T", "file conforms to FITS standard")
		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 0, "number of array dimensions")
		finish_header()
		write("XTENSION", repr("BINTABLE"), "binary table extension")

		write("BITPIX", 8, "number of bits per data pixel")
		write("NAXIS", 2, "number of array dimensions")
		write("NAXIS1", dataset.byte_size(selection=selection), "length of dim 1")
		write("NAXIS2", 1, "length of dim 2")

		write("TFIELDS", len(dataset.column_names), "number of columns")
		for i, column_name in enumerate(dataset.column_names):
			i += 1 # 1 based index
			column = dataset.columns[column_name]
			write("TTYPE%d" % i, repr(str(column_name)), "column name %i" % (i))
			numpy_type_name = column.dtype.descr[0][1][1:] # i4, f8 etc
			fits_type = astropy.io.fits.column.NUMPY2FITS[numpy_type_name]
			# TODO: support rank1 arrays
			write("TFORM%d" % i , repr("{length}{type}".format(length=len(dataset), type=fits_type)), "")
			write("TDIM%d" % i, repr("(1,{length})".format(length=len(dataset))), "")

		finish_header()

		for i, column_name in enumerate(dataset.column_names):
			column = dataset.columns[column_name]
			numpy_type_name = column.dtype.descr[0][1][1:] # i4, f8 etc
			fits_type = astropy.io.fits.column.NUMPY2FITS[numpy_type_name]
			chunk_size = 1024**2 # 1 meg at a time
			chunks = int(math.ceil(len(dataset)/float(chunk_size)))
			for i in range(chunks):
				i1 = i * chunk_size
				i2 = min(len(dataset), (i+1) * chunk_size)
				data_big_endian = column[i1:i2].astype(">" + numpy_type_name)
				f.write(data_big_endian)
			print((f.tell(), f.tell() / 1024**2, "mb", len(dataset)))
			assert i2 == len(dataset)
		finish_data()


def export_fits(dataset, path, column_names=None, shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True):
    """
    :param DatasetLocal dataset: dataset to export
    :param str path: path for file
    :param lis[str] column_names: list of column names to export or None for all columns
    :param bool shuffle: export rows in random order
    :param bool selection: export selection or not
    :param progress: progress callback that gets a progress fraction as argument and should return True to continue,
            or a default progress bar when progress=True
    :param: bool virtual: When True, export virtual columns
    :return:
    """
    if shuffle:
        random_index_name = "random_index"
        while random_index_name in dataset.get_column_names():
            random_index_name += "_new"

    column_names = column_names or dataset.get_column_names(virtual=virtual, strings=True)
    logger.debug("exporting columns(fits): %r" % column_names)
    N = len(dataset) if not selection else dataset.selected_length(selection)
    data_types = []
    data_shapes = []
    ucds = []
    units = []
    for column_name in column_names:
        if column_name in dataset.get_column_names(strings=True, virtual=False):
            column = dataset.columns[column_name]
            shape = (N,) + column.shape[1:]
            dtype = column.dtype
            if dataset.is_string(column_name):
                max_length = dataset[column_name].apply(lambda x: len(x)).max(selection=selection)
                dtype = np.dtype('S'+str(int(max_length)))
        else:
            dtype = np.float64().dtype
            shape = (N,)
        ucds.append(dataset.ucds.get(column_name))
        units.append(dataset.units.get(column_name))
        data_types.append(dtype)
        data_shapes.append(shape)

    if shuffle:
        column_names.append(random_index_name)
        data_types.append(np.int64().dtype)
        data_shapes.append((N,))
        ucds.append(None)
        units.append(None)
    else:
        random_index_name = None

    # TODO: all expressions can have missing values.. how to support that?
    null_values = {key: dataset.columns[key].fill_value for key in dataset.get_column_names() if dataset.is_masked(key) and dataset.data_type(key).kind != "f"}
    empty(path, N, column_names, data_types, data_shapes, ucds, units, null_values=null_values)
    if shuffle:
        del column_names[-1]
        del data_types[-1]
        del data_shapes[-1]
    dataset_output = vaex.astro.fits.FitsBinTable(path, write=True)
    df_output = vaex.dataframe.DataFrameLocal(dataset_output)
    vaex.export._export(dataset_input=dataset, dataset_output=df_output, path=path, random_index_column=random_index_name,
            column_names=column_names, selection=selection, shuffle=shuffle,
            progress=progress, sort=sort, ascending=ascending)
    dataset_output.close()
